// parse arguments

var conf = {
	'port':     8081,
	'listen': '127.0.0.1',
	'hand_size': 10,
	'socket-debug': false
};

var optinfo = require('node-getopt').create([
	['h', 'help',          'Display this help'],
	['p', 'port=PORT',     'Port'],
	['l', 'listen=IP',     'IP address to bind to'],
	['',  'hand-size=NUM', 'Hand size'],	
	['',  'socket-debug',  'Display socket.io debug messages']
	
]).bindHelp().parseSystem();

for (var i in optinfo.options) {
	conf[i] = optinfo.options[i]; // overwrite default
}

//

var cards = require('./static/cards.js').cards;

var http_server = function(req, res) {
	var accept = {
		'/'                     : ['/index.htm', 'text/html'],
		'/app.js'               : [false, 'text/javascript'],
		'/jquery-1.11.0.min.js' : [false, 'text/javascript'],
		'/cards.js'             : [false, 'text/javascript']
	};
	if (req.url in accept) {
		var path = (accept[req.url][0] === false ? req.url : accept[req.url][0]);
		fs.readFile(
			__dirname + '/static' + path,
			function(err, data) {
				if (err) {
					console.log('error serving ' + path);
					res.writeHead(500);
					res.end();
				} else {
					console.log('serving ' + path);
					res.writeHead(200, {
						'Content-Length': data.length,
						'Content-Type'  : accept[req.url][1]
					});
					res.end(data);
				}
			}
		);
	} else {
		console.log('bad request ' + req.url);
		res.writeHead(404);
		res.end('Not Found');
	}
};

//

/**
 * Shuffle an array. Returns new array by value
 * 
 * @param {Array} arr
 * @returns {Array}
 */
var array_shuffle = function(arr) {
	var copy = Array.prototype.slice.call(arr);
	
	for (var i = 0, e = arr.length; i !== e; ++i) {
		var idx = Math.floor(Math.random() * e);
		
		var tmp   = copy[idx];
		copy[idx] = copy[i];
		copy[i]   = tmp;
	};
	
	return copy;
};

/**
 * Compare two arrays elementwise.
 * 
 * @param {Array} a
 * @param {Array} b
 * @returns {Boolean}
 */
var array_equal = function(a, b) {
	if (a.length === b.length) {
		for (var i = 0, e = a.length; i !== e; ++i) {
			if (a[i] !== b[i]) return false;
		}
		return true;
	}
	return false;
};

/**
 * Retrieve object values (analogous to Object.keys())
 * 
 * @param {Object} obj
 * @returns {Array}
 */
var object_values = function(obj) {
	var ret = [];
	for (var key in obj) {
		ret.push( obj[key] );
	}
	return ret;
};

//

var games = {};

var generate_game_id = function() {
	return ( Math.random().toString(36).substr(2, 5).toUpperCase() );
};

var game_new = function(name) {
	var key = generate_game_id();
	games[key] = {
		'name': name,
		'players': {
			/*
			'name': {
				'socket': $socket,
				'score': $score,
				'hand': [white cards...]
			}
			*/
		},
		'czar': '',
		'question': 0,
		'submissions': {}, // current round submissions
		'deck': array_shuffle(Object.keys( cards.white )) // keys only, not text
	};
	
	return key;
};

var game_draw = function(key, num) {
	var ret = [];
	while(num-- > 0) {
		ret.push( games[key]['deck'].pop() );		
	};
	
	return ret;
};

var game_newround = function(key) {
	var czar = '';
	var question = 0;
	
	if (Object.keys(games[key]['players']).length) {
		var players = Object.keys(games[key]['players']);
		czar = players[Math.floor(Math.random() * players.length)];
		
		question = Math.floor(Math.random() * cards.black.length); // key only, not text
	};
	
	games[key]['czar'] = czar;
	games[key]['question'] = question;
	games[key]['submissions'] = {};
	
	// Top up hands for all players
	for (var p in games[key]['players']) {
		var hand = games[key]['players'][p]['hand'];
		if (hand.length < conf.hand_size) {
			games[key]['players'][p]['hand'] = hand.concat( game_draw(key, conf.hand_size - hand.length));
			games[key]['players'][p]['socket'].emit('hand', games[key]['players'][p]['hand']);
		}
	}
	
	game_sendall( key, 'room_info', game_getinfo(key) );
	
	console.log('[' + key + '] New round, '+games[key]['deck'].length+' in deck');
};

var game_getinfo = function(key) {
	if (key in games) {
		var ret = {
			'name': games[key]['name'],
			'players': {},
			'czar': games[key]['czar'],
			'question': games[key]['question']
		};
		
		for (var i in games[key]['players']) {
			ret['players'][i] = games[key]['players'][i]['score'];
		};
		
		return ret;
	
	} else {
		return false;
	}
};

var game_sendall = function(key, message, data) {
	if (key in games && Object.keys(games[key]['players']).length) {
		for(var i in games[key]['players']) {
			games[key]['players'][i]['socket'].emit( message, data );
		}
	}
};

var game_addplayer = function(key, name, socket) {
	if (games[key]['players'][name]) {
		return false;
		
	} else {
		game_sendall(key, 'player_join', name);
		
		games[key]['players'][name] = {
			'score': 0,
			'socket': socket,
			'hand': game_draw(key, conf.hand_size)
		};
		
		return true;
		
	}
};

//

var getroomlist = function() {
	var ret = [];
	for (var i in games) {
		ret.push({
			'key': i,
			'name': games[i]['name'],
			'playerct': Object.keys(games[i]['players']).length
		});
	};
	return ret;
};

// we SEND: log, room_set, room_info, player_join
// we RECV: game_new, game_join


var sio_server = function(socket) {
	socket.emit('hi');
	
	socket.on('listrooms', function() {
		socket.emit('room_list', getroomlist());
	});
	
	socket.on('game_new', function(data) {
		var key = game_new(data.name);
		socket.emit('room_set', key);
	});
	
	socket.on('game_join', function(data) {
		var key = data.key;
		var player_name = data.name;
		
		if (key in games) {
			if (game_addplayer(key, player_name, socket)) {
				socket.emit('room_info', game_getinfo(key)); // now includes self
				socket.emit('hand', games[key]['players'][player_name]['hand']);
				
				if (Object.keys(games[key]['players']).length === 1) {
					game_newround(key); // player 1
				}
				
				// cache player's current game in socket object
				socket.set('playerinfo', {'key': key, 'name': player_name});
				
			} else {
				socket.emit('crash', 'Error joining room (nick in use?)');
			}
		
		} else {
			socket.emit('crash', 'Error joining room (invalid room)');
		}
		
	});
	
	socket.on('submit', function(data) {
		// data is an array of card indices
		// Identify which game this user is in
		
		socket.get('playerinfo', function(err, playerinfo) {
			//console.log([err, playerinfo]);
			
			var name = playerinfo.name;
			var gamekey = playerinfo.key;
			
			// TODO: Verify that the cards were all present in this user's hand
			// TODO: Verify that this is the right socket for this user
			// TODO: Verify that this is not a double-submit

			// Cache submission
			games[gamekey]['submissions'][name] = data;

			// Remove cards from player's serverside hand
			for (var i in data) {
				var idx = games[gamekey]['players'][name]['hand'].indexOf( data[i] ); // find in hand
				games[gamekey]['players'][name]['hand'].splice(idx, 1); // remove
				
				// add to bottom of draw pile
				games[gamekey]['deck'].unshift(i);
			};

			// If this was the final submission, lapse into czar selection
			var submitting_players_ct = Object.keys(games[gamekey]['players']).length - 1;
			var submissions_so_far    = Object.keys(games[gamekey]['submissions']).length;
			
			console.log('['+gamekey+'] Submission '+submissions_so_far+' / '+submitting_players_ct);
			
			if (submissions_so_far === submitting_players_ct) {
				console.log('['+gamekey+'] Czar select');
				game_sendall(gamekey, 'czar_select', object_values( games[gamekey]['submissions'] ));
			}
		});
		
	});
	
	socket.on('czar_chooses', function(data) {
		// data is a game submission (array of answers)
		
		// Identify which game this user is in
		socket.get('playerinfo', function(err, playerinfo) {
			var name = playerinfo.name;
			var gamekey = playerinfo.key;
			
			// TODO: Verify the selection is valid
			// Find which user submitted this entry
			var winner = null;
			for (var i in games[gamekey]['submissions']) {
				if (array_equal(data, games[gamekey]['submissions'][i])) {
					winner = i;
					break;
				}
			}
			
			if (winner === null) {
				// TODO: Dramatic failure
			}
			
			// Allocate points
			games[gamekey]['players'][i]['score'] += 1;

			// New round
			game_newround(gamekey);
		});
		
		
	});
	
	socket.on('disconnect', function(/*data*/) {
		// emitted by platform.
		
		socket.get('playerinfo', function(err, playerinfo) {
			if (! playerinfo) return; // this connection hadn't joined a game yet
			
			var name = playerinfo.name;
			var gamekey = playerinfo.key;
			
			if (name in games[gamekey]['players']) { // sanity check

				// Move hand back to deck
				for (var i = 0; i < games[gamekey]['players'][name]['hand'].length; ++i) {
					games[gamekey]['deck'].unshift( games[gamekey]['players'][name]['hand'][i] );
				}

				// Shuffle deck
				games[gamekey]['deck'] = array_shuffle(games[gamekey]['deck']);

				// Drop player
				delete games[gamekey]['players'][name];

				// Notify other players
				if (games[gamekey]['czar'] === name) {
					// new round
					game_newround(gamekey);
					game_sendall(gamekey, 'log', 'Czar "'+name+'" quit, starting new round.');

				} else {
					game_sendall(gamekey, 'log', 'Player "'+name+'" quit.');
				}
				
			}
		});
	});
};

//

var http = require('http').createServer(http_server);
var io   = require('socket.io').listen(http);
var fs   = require('fs');

if (! conf['socket-debug']) {
	io.set('log level', 1);
}

io.sockets.on('connection', sio_server);
console.log('Registered socket.io.');

http.listen(conf.port, conf.listen);
console.log('Web server started on http://'+conf.listen+":"+conf.port+"/");


var do_maintenance = function() {
	for (var i in games) {
		if (Object.keys(games[i]['players']).length === 0) {
			console.log("["+i+"] Pruning game");
			delete games[i];
		}
	}
};

var _maintenance = setInterval(do_maintenance, 5 * 60 * 1000);